'''文件操作'''


if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


import os
from os.path import dirname, basename
import subprocess
import tkinter as tk
from tkinter.simpledialog import askstring


class FileManager(tk.Menu):
    def __init__(self, parent):
        text = parent.text
        root = parent.root
        io = parent.io

        tk.Menu.__init__(self, root, tearoff=0)
        n = io.filename or os.getcwd() + '\\' # 当新建文件时 io.filename==None 尾部加分隔符避免获取路径出错
        text_insert   = lambda s: (lambda: text.insert('insert', s))
        clipboard_set = lambda s: (lambda: (root.clipboard_clear(), root.clipboard_append(s)))

        self.add_command(label='Copy Fullname', command=clipboard_set(n))
        self.add_command(label='Copy Filename', command=clipboard_set(basename(n)))
        self.add_command(label='Copy Dirname',  command=clipboard_set(dirname(n)))
        self.add_separator()
        self.add_command(label='Insert Fullname', command=text_insert("r'%s'"%n))
        self.add_command(label='Insert Filename', command=text_insert("'%s'"%basename(n)))
        self.add_command(label='Insert Dirname',  command=text_insert("r'%s'"%dirname(n)))
        self.add_separator()
        self.add_command(label='Open in Explorer', command=lambda: subprocess.Popen('explorer /select, "%s"'%n))
        self.add_command(label='Open in CMD',      command=lambda: os.system('start'))
        # self.add_command(label='Run in CMD',       command=lambda: os.system('start "%s"'%io.filename)) # TODO
        self.add_separator()
        self.add_command(label='Rename File', command=lambda: self.Rename(text, io))
        self.add_command(label='Reload File', command=lambda: self.Reload(io))

        text.bind('<F2>', lambda e: self.Rename(text, io))

        parent.amenu.insert_cascade(3, label='File Manager', menu=self)

    def Rename(self, text, io):
        basename2 = basename(io.filename)
        filename2 = askstring('Rename', 'Input new filename:', initialvalue=basename2, parent=text) # 允许输入相对路径
        if filename2:
            if filename2[-3:].lower() != '.py':
                filename2 += '.py'
            dirname2 = dirname(filename2)
            if dirname2: # 只输入文件名的时候路径为 ''
                os.makedirs(dirname2, exist_ok=True)
            os.rename(io.filename, filename2) # TODO 切换目录时工作路径未变化
            io.set_filename(os.path.abspath(filename2)) # TODO save状态变化

    def Reload(self, io): # TODO 重载记忆位置
        io.loadfile(io.filename)  # TODO rename也可以用这个方法



'''智能选取'''

# TODO shell中的stdout无法匹配

if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


import re

sp = lambda c: eval(c.replace('.',',')) # Good!


def Select(text, c1, c2):
    text.tag_remove('sel', '1.0', 'end')
    text.tag_add('sel', c1, c2)
    text.mark_set('insert', c1)


def FixTextSelect(root):
    tk = root.tk
    tk.eval('bind Text <Double-1> {catch {%W mark set insert sel.first}}') # See: "~\tcl\tk8.6\text.tcl"


def CurrentInTag(text, tag):
    cur = text.index('current')
    pre = text.tag_prevrange(tag, cur)
    nex = text.tag_nextrange(tag, cur)
    if pre and (sp(pre[0]) <= sp(cur) < sp(pre[1])):
        return pre
    if nex and (sp(nex[0]) <= sp(cur) < sp(nex[1])):
        return nex


def FindParen(s, c1='(', c2=')'):
    lv = 0
    for n, c in enumerate(s):
        if c in c1:
            lv += 1
        elif c in c2:
            lv -= 1
        if lv == 0:
            return n


def MatchSpan(r, s, n):
    # TODO 当无匹配时的返回最近结果
    for m in re.finditer(r, s):
        if m.start() <= n <= m.end():
            return m.span()


def Selecting(e):
    text = e.widget
    text.tag_remove('hit', '1.0', 'end')

    comment_or_string = CurrentInTag(text, 'COMMENT') or CurrentInTag(text, 'STRING')
    if comment_or_string: # TODO 注释前的空格
        c1, c2 = comment_or_string
        text.mark_set('insert', c1)
        text.tag_remove('sel', '1.0', 'end')
        text.tag_add('sel', c1, c2)
        return

    cur = text.index('current') # 当用insert时光标位置为自动选区的最开始
    # print(cur)
    ln, col = sp(cur)

    ss = text.get(cur+' linestart', 'end').split('\n')
    line = ss[0]
    indent = re.match(r' *', line).end()
    empty = not line.split('#')[0].strip()

    if col <= indent: # cursor postion
        for row in range(1, len(ss)):
            begin = ss[row] if empty else ss[row][:indent+1] # empty代表无穷缩进
            if begin.split('#')[0].strip() != '' or empty: # 起始时是empty则只选中一行
                break
        Select(text, 'current linestart', 'current linestart+%dl' % row)

    else:
        c = text.get('current')
        if c == '\n':
            Select(text, 'current linestart', 'current linestart+1l')

        elif c == ' ':
            p1, p2 = MatchSpan(r' +', line, col)
            Select(text, 'current linestart+%dc' % p1, 'current linestart+%dc' % p2)

        elif c in '([{':
            # idle内自建的方法太复杂 idlelib.hyperparser.HyperParser.get_surrounding_brackets
            # TODO 字符串内、转义引号的问题
            c1 = c
            c2 = ')]}'['([{'.index(c1)]
            s = text.get('current', 'end')
            n = FindParen(s, c1, c2)
            Select(text, 'current', 'current+%dc' % (n + 1))

        elif c in ')]}':
            c1 = c
            c2 = '([{'[')]}'.index(c1)]
            s = text.get('1.0', 'current+1c')
            n = FindParen(reversed(s), c1, c2)
            Select(text, 'current-%dc' % n, 'current+1c')

        elif c == ',':
            # TODO 两个括号的问题：(a, b())
            s = text.get('current', 'end')
            for n, c in enumerate(s[1:]):
                if c in ',{[()]}\n':
                    Select(text, 'current', 'current+%dc' % (n + 1))
                    break

        elif re.match(r'\w', c):
            # TODO 换个颜色（不是hit）（参考：notepad++是深绿和浅绿）
            # TODO 选中后取消
            # TODO 一行中第一个word如果是一个字符无法选中（命中另一个规则）
            p1, p2 = MatchSpan(r'\w+', line, col)
            Select(text, 'current linestart+%dc' % p1, 'current linestart+%dc' % p2)
            word = line[p1:p2]
            s = text.get('1.0', 'end')
            for m in re.finditer(r'\b%s\b' % word, s):
                p1, p2 = m.span()
                text.tag_add('hit', '1.0+%dc' % p1, '1.0+%dc' % p2)

        elif re.match(r'\W', c):
            # 改善“s.split('\n')[:ln]”会匹配”')[:“的糟糕匹配体验
            p1, p2 = MatchSpan(r'[^\w\(\)\[\]\{\}\'" ]+', line, col) # 不"粘"括号空格和引号
            Select(text, 'current linestart+%dc' % p1, 'current linestart+%dc' % p2)


class SmartSelect:
    def __init__(self, parent):
        FixTextSelect(parent.root)
        parent.text.bind('<Double-Button-1>', Selecting)



'''记录备份'''


if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


import os


def UniqueFile(file):
    root = os.path.splitext(file)[0]
    n = 1
    while os.path.exists(file):
        n += 1
        file = '%s.%d.pybak'%(root, n)
    return file


class SaveArchive:
    def __init__(self, parent):
        return  # TODO too much backup files.

        self.io = parent.io
        parent.after_save.append(self.Saving)

    def Saving(self):
        self.io.writefile(UniqueFile(self.io.filename))



# TODO 增加右键菜单和拖拽启动打包

import os
import glob
import time
import zipfile


def mark(target):
    tt = time.strftime('.%Y%m%d_%H%M%S')
    base, ext = os.path.splitext(target)
    os.rename(target, base + tt + ext)


def compress(paths, except_key=()):
    save_name = os.path.splitext(paths[0])[0] + time.strftime('.%Y%m%d_%H%M%S.zip')
    zip = zipfile.ZipFile(save_name, 'w', zipfile.ZIP_DEFLATED)
    for path in paths:
        if os.path.isfile(path):
            zip.write(path, path)
        else:
            for file in glob.iglob('%s/**' % path, recursive=True):
                if all(key not in file for key in except_key):
                    zip.write(file, file)
    zip.close()


# mark('idlealib.zip')

lst = [file for file in os.listdir() if os.path.isfile(file) and not file.startswith('test') and not file.endswith('.zip')]
print(lst)

compress(('idlealib', *lst),
         except_key=('__pycache__',))



# Copyright (c) 2021 Lishixian (znsoooo). All Rights Reserved.
#
# Distributed under MIT license.
# See file LICENSE for detail or copy at https://opensource.org/licenses/MIT


'''
运行__main__.py获得一个加载所有插件的IDLE-Advance的示例文件。
运行__init__.py获得一个打开自身脚本并加载所有插件的editor的例子。
分别运行idlealib目录下的扩展文件，可以得到一个打开自身的editor或shell的例子。
如果需要停用部分扩展，将对应的脚本移出目录后重启IDLE即可。
'''


import os

import idlelib.calltip

EXTENSIONS = []

# idlelib.calltip._MAX_LINES = 999
# idlelib.calltip._MAX_COLS = 999


class Calltip(idlelib.calltip.Calltip):
    def __init__(self, editwin=None):
        super().__init__(editwin)
        editwin.ctip = self # make hook

    # def fetch_tip(self, expression):
    #     self.expression = expression
    #     self.argspec = super().fetch_tip(expression)
    #     return self.argspec


idlelib.calltip.Calltip = Calltip



import keyword
import idlelib.autocomplete


class MyAutoComplete(idlelib.autocomplete.AutoComplete):
    def fetch_completions(self, what, mode):
        ret = super().fetch_completions(what, mode)
        if mode == idlelib.autocomplete.COMPLETE_ATTRIBUTES and what == '':
            for lst in ret[:2]:
                lst.extend(v for v in keyword.kwlist if v not in lst) # `None/True/False` are repetitive.
                lst.sort()
        return ret


idlelib.autocomplete.AutoComplete = MyAutoComplete




def FixPath():
    # To fix open shell (call run() without sys.argv) or self.load_extension(name) will not work.
    import sys
    path = os.path.dirname(__file__)
    if path not in sys.path:
        sys.path.insert(0, path)


import idlelib.editor
from idlelib.editor import EditorWindow
from idlelib.mainmenu import menudefs

# editor.EditorWindow -> pyshell.PyShellEditorWindow
# editor.EditorWindow -> outwin.OutputWindow -> pyshell.PyShell


mymenudef = ('advance', [])

menudefs.append(('advance', []))


class MyEditorWindow(EditorWindow):
    def __init__(self, *args):
        if ('advance', 'Advance') not in self.menu_specs:
            self.menu_specs.append(('advance', 'Advance'))

        EditorWindow.__init__(self, *args)

        self.amenu = self.menudict['advance']
        self.make_rmenu() # make "self.rmenu"

        self.before_copy = []
        self.after_save  = []
        self.after_close = []

        text = self.text
        text.bind('<<save-window>>', self.save) # fix all event handle in this class.
        text.bind('<F12>', self.Test)

        self.recent_files_menu['postcommand'] = self.update_recent_files_list # fix list not refresh when open another IDLE.

        self.load_adv_extensions()

    def cut(self, event):
        [f() for f in self.before_copy] # same as `copy`
        return super().cut(event)

    def copy(self, event):
        [f() for f in self.before_copy]
        return super().copy(event)

    def _close(self):
        print('handle with edit _close:', self.io.filename)
        super()._close()
        # raise # TODO 是否还有别的方法阻止清空剪切

    def close(self):
        # "<<close-window>>"事件不命中点击窗口关闭事件
        print('handle with edit close:', self.io.filename)
        [f() for f in self.after_close]
        super().close()

    def save(self, e):
        self.io.save(e)
        [f() for f in self.after_save]

    def add_adv_menu(self, label, sub, index='end', sp=False):
        menu = self.menudict['advance']
        if sp and menu.index('end') is not None:
            menu.insert_separator(index)
        if callable(sub):
            menu.insert_command(index, label=label, command=sub)
        else:
            menu.insert_cascade(index, label=label, menu=sub)

    def load_adv_extensions(self):
        for file in EXTENSIONS:
            name, ext = os.path.splitext(os.path.basename(file))
            if ext == '.py' and name not in ['__init__', '__main__']:
                try:
                    self.load_extension(name) # TODO 支持任意位置文件导入
                except:
                    print('Failed to import IDLE-Adv extension: %s' % name)
                    import traceback
                    traceback.print_exc()

        menu = self.menudict['advance']
        if menu.type('end') == 'separator':
            menu.delete('end')

    def Test(self, e):
        print('editor on test')
        print('mark_names:', self.text.mark_names())
        print('tag_names:', self.text.tag_names())
        print('functions:', ' '.join(v for v in dir(self.text) if 'tag' in v or 'mark' in v))


idlelib.editor.EditorWindow = MyEditorWindow
from idlelib.pyshell import main  # must after hot patch


def run(filename=None, exts=()):
    FixPath()
    EXTENSIONS.extend(exts)
    if not EXTENSIONS:
        EXTENSIONS.extend(file for file in os.listdir(os.path.dirname(__file__)))
    if filename:
        import sys
        sys.argv.append(filename) # Idea from "~\Lib\tkinter\__main__.py"
    main()


def test_editor(script_file):
    run(script_file, [script_file])


def test_shell(script_file):
    run(None, [script_file])


if __name__ == '__main__':
    run(__file__)


'''
TODO 参考历史文件的打开方法，用于拖拽打开和恢复打开文件
def __recent_file_callback(self, file_name):
    def open_recent_file(fn_closure=file_name):
        self.io.open(editFile=fn_closure)
    return open_recent_file

# 打开文件并定位位置（恢复打开）
outwin.py:
self.flist.gotofileline(filename, lineno)
'''

# TODO 当不在instance_dict中将不会刷新（新启动的idle线程）
'''
def update_recent_files_list(self, new_file=None):
    ...
    # for each edit window instance, construct the recent files menu
    for instance in self.top.instance_dict:
        menu = instance.recent_files_menu
'''



'''文本比较'''


if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


import os
import difflib
import webbrowser
import tkinter as tk
from idlelib.textview import ViewWindow, Button  # Button is ttk.Button


# TODO 支持选择py或全部格式文件
# TODO 优先比对已经打开的文件（列表）

def Comparing(parent, file1, file2):  # TODO 移植到不依赖输入，支持任意文件接口
    with open(file1, encoding='u8') as f:
        ss1 = f.read().split('\n')
    with open(file2, encoding='u8') as f:
        ss2 = f.read().split('\n')
    title = 'Different between %s and %s' % (os.path.basename(file1), os.path.basename(file2))  # TODO 我之前时怎么命名的？

    d = difflib.Differ()
    ss3 = d.compare(ss1, ss2)
    text = '\n'.join(ss3)
    dlg = ViewWindow(parent, title, text, wrap='none', _utest=True)
    dlg.viewframe.button_ok.forget()  # ok按钮用的是ViewFrame里的控件!

    def show_html():
        d = difflib.HtmlDiff()
        with open(title + '.html', 'w', encoding='u8') as f:
            f.write(d.make_file(ss1, ss2))
        webbrowser.open(title + '.html')

    toolbar = tk.Frame(dlg)
    Button(toolbar, text='HTML', command=show_html).pack(side='left')
    Button(toolbar, text='Close', command=dlg.viewframe.ok).pack(side='left')
    toolbar.pack()

    dlg.wait_window()


class CompareFile:
    def __init__(self, parent):
        self.text = parent.text
        self.io = parent.io

        parent.add_adv_menu('Compare to File', self.OnCompareFile, sp=True)

    def OnCompareFile(self):
        file1 = self.io.filename
        file2 = self.io.askopenfile()
        if file2:
            Comparing(self.text, file1, file2)



'''提取说明文件到readme.md'''

import os
import glob
import imp

os.chdir('..')
for file in glob.iglob('*.py'):
    pkg = imp.load_package(file, file) # every package's `name` should not repeat.
    print('- ' + file)
    print(str(pkg.__doc__).strip())
    print()



'''清空Shell'''


if __name__ == '__main__':
    import __init__
    __init__.test_shell(__file__)


# TODO 增加菜单 'Clear Shell Window'

from idlelib.pyshell import PyShell


class ClearShell:
    def __init__(self, parent):
        if isinstance(parent, PyShell):
            self.text = parent.per.bottom # TODO 研究一下作用原理
            self.text.bind("<<clear-window>>", self.clear_window)
            self.text.bind("<Control-l>", self.clear_window)

    def clear_window(self, event):
        text = self.text
        text.delete('1.0', 'iomark linestart')


# Ref: ~\idlexlib\extensions\ClearWindow.py



import os
import re
import sys


def lc(s):
    ss = s.split('\n')
    cur = '%d.%d' % (len(ss), len(ss[-1]))
    return cur


def FindKey(key, path='lib/idlelib'):
    print('\nFindkey:\n"%s" in "%s":'%(key, path))
    for root, folders, files in os.walk(os.path.join(sys.base_prefix, path)):
        for file in files:
            if file.endswith('.py'):
                with open(os.path.join(root, file), encoding='u8') as f:
                    s = f.read()
                for m in re.finditer(key, s):
                    print(file, lc(s[:m.start()]), m.group().strip())
    print()


def PrintTags(text):
    print(text.tag_names())
    for name in text.tag_names():
        print(name, text.tag_ranges(name))


if __name__ == '__main__':
    # FindKey('hello......')

    FindKey('.*Escape.*')

    # FindKey(r'.*\bCopy\b.*')
    # FindKey(r'.*\b\.tcl\b.*', path='lib/tkinter')

    # FindKey(r'.*EditorWindow.*')

    # FindKey(r'.*<<.*>>.*')

    # FindKey(r'FocusIn')
    # FindKey(r'FOCUSIN')

    # FindKey(r".*\bListedToplevel\b.*")
    # FindKey(r".*\bunregister_callback\b.*")
    # FindKey(r".*\bdict\.\b.*")



'''搜索替换工具条'''

# TODO 处理好和纵向滚动条的相对位置问题
# TODO shell中的stdout无法匹配


if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


import re
import tkinter as tk


def SelectSpan(text, span, ins):
    c1, c2 = '1.0+%dc' % span[0], '1.0+%dc' % span[1]
    text.tag_remove('sel', '1.0', 'end')
    text.tag_add('sel', c1, c2)
    text.mark_set('insert', [c1, c2][ins])
    text.see(c1)


def PrepFind(s, pat, repl, isre=False, case=True, word=False):
    if not isre: # TODO self test
        sp = '\\^$.*+?|()[]{}' # "\"需要放在最前面，否则会发生2次替换
        for c in sp:
            pat = pat.replace(c, '\\' + c)
        repl = repl.replace('\\', '\\\\')

##    if not case: # TODO 不区分大小写不应替换原始文本的大小写
##        s = s.lower()
##        pat = pat.lower()

    if word:  # "\b"在"去正则化"之后转换
        pat = r'\b' + pat + r'\b'

    return pat, repl


class ReplaceBar(tk.Frame):
    def __init__(self, parent):
        tk.Frame.__init__(self, parent.text_frame)

        self.show = True

        self.text = parent.text

        # TODO 从idle中获取设置
        self.patvar  = tk.StringVar(self, '')     # search pattern
        self.replvar = tk.StringVar(self, '')     # replace string
        self.revar   = tk.BooleanVar(self, False) # regular expression?
        self.casevar = tk.BooleanVar(self, False) # match case?
        self.wordvar = tk.BooleanVar(self, False) # match whole word?
        self.backvar = tk.BooleanVar(self, False) # search backwards?

        self.patvar .trace('w', self.Setting) # TODO 绑定active事件
        self.replvar.trace('w', self.Setting)
        self.revar  .trace('w', self.Setting)
        self.casevar.trace('w', self.Setting)
        self.wordvar.trace('w', self.Setting)
        self.backvar.trace('w', self.Setting)

        t1 = tk.Entry(self, width=8, textvariable=self.patvar)
        t2 = tk.Entry(self, width=8, textvariable=self.replvar, validatecommand=self.Find)
        tk.Label(self, text='Find:').pack(side='left')
        t1.pack(side='left', fill='x', expand=True)
        tk.Label(self, text='Repl:').pack(side='left')
        t2.pack(side='left', fill='x', expand=True)

        self.tip = tk.Label(self, text=' Match: 0')
        self.tip.pack(side='left')

        tk.Checkbutton(self, text='Cc', variable=self.casevar).pack(side='left')
        tk.Checkbutton(self, text='Wd', variable=self.wordvar).pack(side='left')
        tk.Checkbutton(self, text='Re', variable=self.revar)  .pack(side='left')

        tk.Button(self, relief='groove', text='<<', command=lambda: self.View(0)).pack(side='left')
        tk.Button(self, relief='groove', text='>>', command=lambda: self.View(1)).pack(side='left')
        tk.Button(self, relief='groove', text='Replace', command=self.Replace).pack(side='left')
        tk.Button(self, relief='groove', text='Replace All', command=self.ReplaceAll).pack(side='left')

        self.text.bind('<<replace-bar-show>>', self.Flip)
        self.text.event_add('<<replace-bar-show>>', '<Key-Escape>') # add event but not clear exist bindings.
        t1.bind('<Escape>', self.Flip)
        t2.bind('<Escape>', self.Flip)

        self.Flip(-1)

    def Flip(self, evt):
        self.show = not self.show
        if self.show:
            self.pack(fill='x', side='bottom')
            self.Find()
        else:
            self.forget()
            self.text.tag_remove('hit', '1.0', 'end')
            self.text.focus()

    def Setting(self, *args):
        self.Find()

    def Find(self):
        self.text.tag_remove('hit', '1.0', 'end')
        self.tip.config(text=' Match: 0')

        pat = self.patvar.get()
        if not pat:
            return

        # s = self.text.get('sel.first', 'sel.last')
        s = self.text.get('1.0', 'end')

        repl = self.replvar.get()
        back = self.backvar.get()

        pat, repl = PrepFind(s, pat, repl, self.revar.get(), self.casevar.get(), self.wordvar.get())
        matchs = [m.span() for m in re.finditer(pat, s)]
        self.tip.config(text=' Match: %d' % len(matchs))
        for p1, p2 in matchs:
            self.text.tag_add('hit', '1.0+%dc' % p1, '1.0+%dc' % p2)

        return matchs, s, pat, repl

    def Highlight(self):
        # TODO 未完成
        matchs = [m.span() for m in re.finditer(pat, s)]
        self.tip.config(text=' Match: %d' % len(matchs))
        for p1, p2 in matchs:
            self.text.tag_add('hit', '1.0+%dc' % p1, '1.0+%dc' % p2)

    def View(self, next):
        '''next: 1 -> forward, 0 -> backward'''
        # TODO 移动光标到选区边缘后继续查找
        self.backvar.set(not next)
        matchs, s, pat, repl = self.Find()
        if matchs:
            ins = len(self.text.get('1.0', 'insert')) # cursor offset
            now = sorted([p1 for p1, p2 in matchs] + [ins]).index(ins) - 1
            new = (now+next) % len(matchs)
            SelectSpan(self.text, matchs[new], next)
            self.tip.config(text=' Match: %d/%d' % (new+1, len(matchs)))

    def Replace(self):
        next = not self.backvar.get()
        if not self.text.get('sel.first', 'sel.last'):
            # TODO 如果选区未完全匹配表达式也不替换
            self.View(next)
            return # 第一次只匹配不替换
        self.ReplaceAll()
        self.View(next)

    def ReplaceAll(self):
        ss = self.text.get('sel.first', 'sel.last')
        if not ss:
            self.text.tag_add('sel', '1.0', 'end')

        s1 = self.text.get('sel.first', 'sel.last')
        matchs, s, pat, repl = self.Find()
        s2 = re.sub(pat, repl, s1)
        self.text.delete('sel.first', 'sel.last')
        self.text.insert('insert', s2) # TODO 光标移动到了最后
        # TODO 替换选中的部分


if __name__ == '__main__':
    s = bytes(range(128)).decode()
    pat, repl = PrepFind('', s, s)
    print('PrepFind Test Patt:', s == re.sub(pat, '', s))
    print('PrepFind Test Repl:', s == re.sub('1', repl, '1'))



'''窗口操作'''


if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


import os
import tkinter as tk


class WindowManager(tk.Menu):
    def __init__(self, parent):
        tk.Menu.__init__(self, parent.menubar, tearoff=0)
        self.parent = parent
        self.MakeMenu()
        self.Binding()
        parent.amenu.insert_cascade(3, label='Window Manager', menu=self)

    def GetFileList(self):
        filename = self.parent.io.filename
        op_list = self.parent.top.instance_dict.values()
        with open(self.parent.recent_files_path, encoding='u8') as f:
            rf_list = f.read().splitlines()
        now = -1
        opened = []
        closed = []
        for n, file1 in enumerate(rf_list):
            if os.path.samefile(file1, filename):
                now = n
            # 大小写斜杠不同导致不同字符串但是可能是同一个文件 # TODO os.path.normcase(...) Ref: idlelib.filelist.open
            status = any(os.path.samefile(file1, file2) for file2 in op_list)
            if status:
                opened.append(file1)
            else:
                closed.append(file1)
        return now, opened, closed

    def OnRestore(self, *args): # TODO 恢复窗口时恢复记忆位置
        now, opened, closed = self.GetFileList()
        if closed: # not empty
            self.parent.io.open(editFile=closed[0])

    def OnPrev(self, *args):
        now, opened, closed = self.GetFileList()
        new = (now + 1) % len(opened)
        self.parent.io.open(editFile=opened[new])

    def OnNext(self, *args):
        now, opened, closed = self.GetFileList()
        new = (now - 1) % len(opened)
        self.parent.io.open(editFile=opened[new])

    def OnClose(self, *args): # TODO 多次恢复之后检查关闭后激活下一个的循环顺序问题
        self.OnPrev()
        self.parent._close() # TODO 没有保存!! event: '<<close-all-windows>>', '<<close-window>>'

    def OnCloseAll(self, *args):
        for win in list(self.parent.top.instance_dict): # list the dict avoid dict changed during run.
            win._close()

    def MakeMenu(self):
        self.add_command(label='Restore Window', command=self.OnRestore)
        self.add_separator()
        self.add_command(label='Prev Window', command=self.OnPrev)
        self.add_command(label='Next Window', command=self.OnNext)
        self.add_separator()
        self.add_command(label='Close Window', command=self.OnClose)
        self.add_command(label='Close All Windows', command=self.OnCloseAll)

    def Binding(self):
        text = self.parent.text
        text.bind('<Control-Shift-T>', self.OnRestore)
        text.bind('<Control-Prior>', self.OnPrev)
        text.bind('<Control-Next>', self.OnNext)
        text.bind('<Control-w>', self.OnClose)
        text.bind('<Control-Shift-W>', self.OnCloseAll)



'''匹配成对括号'''

# TODO 快速删除

if __name__ == '__main__':
    import __init__
    __init__.test_editor(__file__)


def Pairing(e):
    # TODO 第一次输入右括号时移动光标但不键入
    # TODO 删除左括号时删除右括号（如果有的话）
    # TODO 有选区时输入括号不清除选区（包括光标位置）
    text = e.widget
    pair = ')]}\'"'['([{\'"'.index(e.char)]
    ss = e.widget.get('sel.first', 'sel.last')
    if ss:
        text.delete('sel.first', 'sel.last')
    text.mark_gravity('insert', 'left')
    text.insert('insert', pair)
    text.insert('insert', ss)
    text.mark_gravity('insert', 'right')


class SmartPairing:
    def __init__(self, parent):
        text = parent.text
        for c in '([{\'"':
            text.bind('<%s>' % c, Pairing)  # '<KeyRelease-%s>'%c
